with Ada.Finalization;
use Ada.Finalization;

generic

  Siz: in Positive; --- largest value of L1,L2

  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Proper_Rounding is <>;

package CG.Spher is

  procedure Initialize_Regge(Parallel: in Boolean);
  procedure GetRegge(L1,L2,L3: in Natural; M1,M2,M3: in Integer; Coe: in out Scalar);
  procedure ProdRegge(L1,L2,L3: in Natural; M1,M2,M3: in Integer; Coe: in out Scalar);

private

  CSize: constant Positive := VSize/8;        --- guessing how many nonzero Regge are needed

  subtype CIndex is Integer range 1 .. CSize; --- enumerates computed nonzero Regge values
  subtype Index2 is Integer range 1 .. VSize; --- range of Indx2, enumerates odered Regge
  type SVec is array(CIndex) of Scalar;       --- for Regge values
  type Controlled_SVec is new Controlled with
    record
      Val: SVec;
    end record;

  LastL3: constant Positive := LMax-Siz;               --- see TSpherics
  Not_Initialized: Boolean := True;
  Count: Natural := 0;                                 --- counts computed Regge
  Pointer: array(Index2) of Integer := (others => -1); --- maps Index2 --> CIndex union {0,-1}
  CS: Controlled_SVec;                                 --- will be deallocated when package exits
  C: SVec renames CS.Val;                              --- Regge values

  --- Regges are C(Pointer(Ind)) where Pointer maps Index2 --> CIndex union {0}
  --- Pointer(Ind)=0 means Regge has a zero value

  protected type Protected_Pointer is
    procedure Check(Ind: in Integer; Ptr: out Integer);
  end Protected_Pointer;

  procedure Compute_Regges(Parallel: in Boolean);

end CG.Spher;
