with Strings;
use Strings;

pragma Elaborate_All (Strings);

package body Regges is

  --------------------------------------------------
  --- N1,N2,N3,M1,M2,M3 are doubled 3j arguments ---
  --------------------------------------------------

  procedure ToRegge(N1,N2,N3: in Natural; M1,M2,M3: in Integer; R: out Regge; Is_Zero: out Boolean) is
    --- N1,N2,N3,M1,M2,M3 are doubled 3j arguments
    K: Integer;
  begin
    Is_Zero := True;
    if (M1+M2+M3)=1 then return; end if;
    if M1=0 and then M2=0 and then ((N1+N2+N3) mod 4) /= 0 then return; end if;
    --- assign and check
    K := N2+N3-N1; if K<0 or else (K mod 2)=1 then return; end if;
    R(I11) := K/2;
    K := N3+N1-N2; if K<0 or else (K mod 2)=1 then return; end if;
    R(I12) := K/2;
    K := N1+N2-N3; if K<0 or else (K mod 2)=1 then return; end if;
    R(I13) := K/2;
    K := N1-M1;    if K<0 or else (K mod 2)=1 then return; end if;
    R(I21) := K/2;
    K := N2-M2;    if K<0 or else (K mod 2)=1 then return; end if;
    R(I22) := K/2;
    K := N3-M3;    if K<0 or else (K mod 2)=1 then return; end if;
    R(I23) := K/2;
    K := N1+M1;    if K<0 or else (K mod 2)=1 then return; end if;
    R(I31) := K/2;
    K := N2+M2;    if K<0 or else (K mod 2)=1 then return; end if;
    R(I32) := K/2;
    K := N3+M3;    if K<0 or else (K mod 2)=1 then return; end if;
    R(I33) := K/2;
    ---
    Is_Zero := False;
  end ToRegge;

  procedure To3j(R: in Regge; N1,N2,N3: out Natural; M1,M2,M3: out Integer) is
    --- this is highly non-unique
    Sum: constant Integer := R(I11)+R(I12)+R(I13);
  begin
    N1 := Sum-R(I11);
    N2 := Sum-R(I12);
    N3 := Sum-R(I13);
    M1 := N1-2*R(I21);
    M2 := N2-2*R(I22);
    M3 := N3-2*R(I23);
  end To3j;

  function Check(R: Regge) return Boolean is
    Sum: constant Integer := R(I11)+R(I12)+R(I13);
  begin
    return     (R(I21)+R(I22)+R(I23))=Sum
      and then (R(I31)+R(I32)+R(I33))=Sum
      and then (R(I11)+R(I21)+R(I31))=Sum
      and then (R(I12)+R(I22)+R(I32))=Sum
      and then (R(I13)+R(I23)+R(I33))=Sum;
  end Check;

  procedure Order(R: in out Regge; ChangeSign: out Boolean) is
    --- R is assumed to be consistent
    R11: Integer renames R(I11);
    R12: Integer renames R(I12);
    R13: Integer renames R(I13);
    R21: Integer renames R(I21);
    R22: Integer renames R(I22);
    R23: Integer renames R(I23);
    R31: Integer renames R(I31);
    R32: Integer renames R(I32);
    R33: Integer renames R(I33);
    Row,Col,Min,Max,K: Integer;
  begin
    ChangeSign := False;
    ------------------------------------------------- find minimum
    Row := 1; Col := 1; Min := R11;
    if R12<Min then           Col := 2; Min := R12; end if;
    if R13<Min then           Col := 3; Min := R13; end if;
    if R21<Min then Row := 2; Col := 1; Min := R21; end if;
    if R22<Min then Row := 2; Col := 2; Min := R22; end if;
    if R23<Min then Row := 2; Col := 3; Min := R23; end if;
    if R31<Min then Row := 3; Col := 1; Min := R31; end if;
    if R32<Min then Row := 3; Col := 2; Min := R32; end if;
    if R33<Min then Row := 3; Col := 3; Min := R33; end if;
    -------------------------------------------------
    if Row=2 then                                  -- swap rows 1 and 2
      K := R11; R11 := R21; R21 := K;
      K := R12; R12 := R22; R22 := K;
      K := R13; R13 := R23; R23 := K;
      ChangeSign := not ChangeSign;
    elsif Row=3 then                               -- swap rows 1 and 3
      K := R11; R11 := R31; R31 := K;
      K := R12; R12 := R32; R32 := K;
      K := R13; R13 := R33; R33 := K;
      ChangeSign := not ChangeSign;
    end if;
    if Col=2 then                                  -- swap columns 1 and 2
      K := R11; R11 := R12; R12 := K;
      K := R21; R21 := R22; R22 := K;
      K := R31; R31 := R32; R32 := K;
      ChangeSign := not ChangeSign;
    elsif Col=3 then                               -- swap columns 1 and 3
      K := R11; R11 := R13; R13 := K;
      K := R21; R21 := R23; R23 := K;
      K := R31; R31 := R33; R33 := K;
      ChangeSign := not ChangeSign;
    end if;
    ------------------------------------------------- find maximum (must be in 1st row or column)
    Row := 1; Col := 1; Max := R11;
    if R12>Max then           Col := 2; Max := R12; end if;
    if R13>Max then           Col := 3; Max := R13; end if;
    if R21>Max then Row := 2; Col := 1; Max := R21; end if;
    if R31>Max then Row := 3; Col := 1; Max := R31; end if;
    -------------------------------------------------
    if Max=Min then return; end if;
    if Row /= 1 then                               -- transpose
      K := R12; R12 := R21; R21 := K;
      K := R13; R13 := R31; R31 := K;
      K := R23; R23 := R32; R32 := K;
      K := Row; Row := Col; Col := K;
    end if;
    if Col /= 2 then                               -- swap columns 2 and 3
      K := R12; R12 := R13; R13 := K;
      K := R22; R22 := R23; R23 := K;
      K := R32; R32 := R33; R33 := K;
      ChangeSign := not ChangeSign;
    end if;
    if R33>R23 then                          -- swap rows 2 and 3
      K := R21; R21 := R31; R31 := K;
      K := R22; R22 := R32; R32 := K;
      K := R23; R23 := R33; R33 := K;
      ChangeSign := not ChangeSign;
    end if;
    ChangeSign := ChangeSign and then ((R11+R12+R13) mod 2)=1;
  end Order;

  function Indx(S,B,T,X,L: Natural) return Positive is
  begin
    return LST_Count(L*LMax2+S*Lmax1+T)+(X-T)*(T-S+1)+(B-S)+1;
  end Indx;

  procedure CheckIndx is
    I: Natural := 0;
  begin
    for L in 0 .. LMax loop
      for S in 0 .. L loop
        for T in S .. (L+S)/2 loop
          for X in T .. L loop
            for B in S .. T loop
              I := I+1;
              if Indx(S,B,T,X,L) /= I then raise Inconsistent; end if;
            end loop;
          end loop;
        end loop;
      end loop;
    end loop;
    if VSize /= I then raise Inconsistent; end if;
  end CheckIndx;

  procedure Indx(R: in Regge; I: out Positive; ChangeSign: out Boolean) is
    --- R is assumed to be consistent
    Rt: Regge := R;
    S: Integer renames Rt(I11);
    B: Integer renames Rt(I22);
    T: Integer renames Rt(I33);
    X: Integer renames Rt(I21);
    L: Integer renames Rt(I12);
  begin
    Order(Rt,ChangeSign);
    if X<T or T<B then raise Inconsistent; end if;
    I := LST_Count(L*LMax2+S*Lmax1+T)+(X-T)*(T-S+1)+(B-S)+1; -- Indx(S,B,T,X,L)
  end Indx;

  procedure Indx(N1,N2,N3: in Natural; M1,M2,M3: in Integer; I: out Natural; ChangeSign: out Boolean) is
    Is_Zero: Boolean;
    R: Regge;
  begin
    ToRegge(N1,N2,N3,M1,M2,M3,R,Is_Zero);
    if Is_Zero then
      I := 0;
    else
      Order(R,ChangeSign);
      declare
        S: Integer renames R(I11);
        B: Integer renames R(I22);
        T: Integer renames R(I33);
        X: Integer renames R(I21);
        L: Integer renames R(I12);
      begin
        if X<T or T<B then raise Inconsistent; end if;
        I := LST_Count(L*LMax2+S*LMax1+T)+(X-T)*(T-S+1)+(B-S)+1; -- Indx(S,B,T,X,L)
      end;
    end if;
  end Indx;

  function Indx(N1,N2,N3: Natural; M1,M2,M3: Integer) return Natural is
    Is_Zero,ChangeSign: Boolean;
    R: Regge;
  begin
    ToRegge(N1,N2,N3,M1,M2,M3,R,Is_Zero);
    if Is_Zero then return 0; end if;
    Order(R,ChangeSign);
    declare
      S: Integer renames R(I11);
      B: Integer renames R(I22);
      T: Integer renames R(I33);
      X: Integer renames R(I21);
      L: Integer renames R(I12);
    begin
      if X<T or T<B then raise Inconsistent; end if;
      return LST_Count(L*LMax2+S*LMax1+T)+(X-T)*(T-S+1)+(B-S)+1; -- Indx(S,B,T,X,L)
    end;
  end Indx;

  procedure OrderedRegge(S,B,T,X,L: in Natural; R: out Regge) is
    --- assuming S <= B <= T <= X <= L, T <= (L+S)/2
  begin
    R := (S,L,X+B-T,X,B,S+L-T,L+B-T,S+X-T,T);
  end OrderedRegge;

  procedure Show(R: in Regge) is
    W: constant Positive := 4;
  begin
    Show0(Strng(R(I11),W) & Strng(R(I12),W) & Strng(R(I13),W));
    Show0(Strng(R(I21),W) & Strng(R(I22),W) & Strng(R(I23),W));
    Show0(Strng(R(I31),W) & Strng(R(I32),W) & Strng(R(I33),W));
  end Show;

  procedure Show(N1,N2,N3: in Natural; M1,M2,M3: in Integer) is
    W: constant Positive := 4;
  begin
    Show0(Strng(N1,W) & "/2" & Strng(N2,W) & "/2" & Strng(N3,W) & "/2" );
    Show0(Strng(M1,W) & "/2" & Strng(M2,W) & "/2" & Strng(M3,W) & "/2" );
  end Show;

  procedure ShowBoth(R: in Regge) is
    N1,N2,N3,M1,M2,M3: Integer;
  begin
    Show(R);
    To3j(R,N1,N2,N3,M1,M2,M3);
    Show(N1,N2,N3,M1,M2,M3);
  end ShowBoth;

  ----------------------------------------------------------
  --- L1,L2,L3,M1,M2,M3 are true 3j arguments (integers) ---
  ----------------------------------------------------------

  procedure Ordered(L1,L2,L3: in Natural; S,B,T,X,L: out Natural; Is_Zero: out Boolean) is
    --- want S <= B <= T <= X <= L
    --- checking for S,B,T,X,L >=0 is automatic
  begin
    if ((L1+L2+L3) mod 2)=0 then
      B := L2; -- want Min{L1,L2,L3}
      T := L3; -- want Mid{L1,L2,L3}
      X := L1; -- want Max{L1,L2,L3}
      if B>T then
        T := L2;
        B := L3;
      end if;
      if T>X then
        X := T;
        T := L1;
        if B>T then
          T := B;
          B := L1;
        end if;
      end if;            --- now B <= T <= X
      if X <= B+T then
        S := B+T-X;      --- smallest Regge entry
        L := T+X-B;      --- largest Regge entry
        Is_Zero := False;
        return;          --- all fine
      end if;
    end if;
    S := 0;
    L := 0;
    Is_Zero := True;
  end Ordered;

  procedure Ordered(L1,L2,L3: in Natural; M1,M2,M3: in Integer; S,B,T,X,L: out Integer; Is_Zero: out Boolean) is
    --- L1,L2,L3,M1,M2,M3 are true 3j arguments (integers)
    --- S,B,T,X,L >=0 is checked here
  begin
    Is_Zero := True;
    if (M1+M2+M3) /= 0 then return; end if;
    S := L2+L3-L1; if S<0 then return; end if; -- R11
    if (S mod 2) /= 0 then return; end if;
    L := L3+L1-L2; if L<0 then return; end if; -- R12
    X := L1-M1;    if X<0 then return; end if; -- R21
    B := L2-M2;    if B<0 then return; end if; -- R22
    T := L3+M3;    if T<0 then return; end if; -- R33
    declare
      R11: Integer renames S;
      R12: Integer renames L;
      R13: Integer;
      R21: Integer renames X;
      R22: Integer renames B;
      R23: Integer;
      R31: Integer;
      R32: Integer;
      R33: Integer renames T;
      Row,Col,Min,Max,Tmp: Integer;
    begin
      R13 := L1+L2-L3; if R13<0 then return; end if;
      R23 := L3-M3;    if R23<0 then return; end if;
      R31 := L1+M1;    if R31<0 then return; end if;
      R32 := L2+M2;    if R32<0 then return; end if;
      ------------------------------------------------- find minimum
      Row := 1; Col := 1; Min := R11;
      if R12<Min then           Col := 2; Min := R12; end if;
      if R13<Min then           Col := 3; Min := R13; end if;
      if R21<Min then Row := 2; Col := 1; Min := R21; end if;
      if R22<Min then Row := 2; Col := 2; Min := R22; end if;
      if R23<Min then Row := 2; Col := 3; Min := R23; end if;
      if R31<Min then Row := 3; Col := 1; Min := R31; end if;
      if R32<Min then Row := 3; Col := 2; Min := R32; end if;
      if R33<Min then Row := 3; Col := 3; Min := R33; end if;
      -------------------------------------------------
      if Row=2 then                                  -- swap rows 1 and 2
        Tmp := R11; R11 := R21; R21 := Tmp;
        Tmp := R12; R12 := R22; R22 := Tmp;
        Tmp := R13; R13 := R23; R23 := Tmp;
      elsif Row=3 then                               -- swap rows 1 and 3
        Tmp := R11; R11 := R31; R31 := Tmp;
        Tmp := R12; R12 := R32; R32 := Tmp;
        Tmp := R13; R13 := R33; R33 := Tmp;
      end if;
      if Col=2 then                                  -- swap columns 1 and 2
        Tmp := R11; R11 := R12; R12 := Tmp;
        Tmp := R21; R21 := R22; R22 := Tmp;
        Tmp := R31; R31 := R32; R32 := Tmp;
      elsif Col=3 then                               -- swap columns 1 and 3
        Tmp := R11; R11 := R13; R13 := Tmp;
        Tmp := R21; R21 := R23; R23 := Tmp;
        Tmp := R31; R31 := R33; R33 := Tmp;
      end if;
      ------------------------------------------------- find maximum (must be in 1st row or column)
      Row := 1; Col := 1; Max := R11;
      if R12>Max then           Col := 2; Max := R12; end if;
      if R13>Max then           Col := 3; Max := R13; end if;
      if R21>Max then Row := 2; Col := 1; Max := R21; end if;
      if R31>Max then Row := 3; Col := 1; Max := R31; end if;
      -------------------------------------------------
      if Max /= Min then
        if Row /= 1 then                               -- transpose
          Tmp := R12; R12 := R21; R21 := Tmp;
          Tmp := R13; R13 := R31; R31 := Tmp;
          Tmp := R23; R23 := R32; R32 := Tmp;
          Tmp := Row; Row := Col; Col := Tmp;
        end if;
        if Col /= 2 then                               -- swap columns 2 and 3
          Tmp := R12; R12 := R13; R13 := Tmp;
          Tmp := R22; R22 := R23; R23 := Tmp;
          Tmp := R32; R32 := R33; R33 := Tmp;
        end if;
        if R33>R23 then                          -- swap rows 2 and 3
          Tmp := R21; R21 := R31; R31 := Tmp;
          Tmp := R22; R22 := R32; R32 := Tmp;
          Tmp := R23; R23 := R33; R33 := Tmp;
        end if;
      end if;
    end;
    if X<T or T<B then raise Inconsistent; end if;
    Is_Zero := False;
  end Ordered;

  procedure Indx2(L1,L2,L3: in Natural; I0: out Natural) is
    --- L1,L2,L3 are true 3j arguments (integers)
    Is_Zero: Boolean;
    S,B,T,X,L: Integer;
  begin
    Ordered(L1,L2,L3,S,B,T,X,L,Is_Zero);
    if Is_Zero then I0 := 0; return; end if;
    I0 := LST_Count(L*LMax2+S*LMax1+T)+(X-T)*(T-S+1)+(B-S)+1; -- Indx(S,B,T,X,L)
  end Indx2;

  procedure Indx2(L1,L2,L3: in Natural; M1,M2,M3: in Integer; I: out Natural) is
    --- L1,L2,L3,M1,M2,M3 are true 3j arguments (integers)
    Is_Zero: Boolean;
    S,B,T,X,L: Integer;
  begin
    Ordered(L1,L2,L3,M1,M2,M3,S,B,T,X,L,Is_Zero);
    if Is_Zero then I := 0; return; end if;
    I := LST_Count(L*LMax2+S*LMax1+T)+(X-T)*(T-S+1)+(B-S)+1; -- Indx(S,B,T,X,L)
  end Indx2;

  procedure Indx2(L1,L2,L3: in Natural; M1,M2,M3: in Integer; I0,I: out Natural) is
    --- L1,L2,L3,M1,M2,M3 are true 3j arguments (integers)
    Is_Zero: Boolean;
    S,B,T,X,L: Integer;
  begin
    Ordered(L1,L2,L3,S,B,T,X,L,Is_Zero);
    if Is_Zero then I0 := 0; I := 0; return; end if;
    I0 := LST_Count(L*LMax2+S*LMax1+T)+(X-T)*(T-S+1)+(B-S)+1; -- Indx(S,B,T,X,L)
                                                              ---
    Ordered(L1,L2,L3,M1,M2,M3,S,B,T,X,L,Is_Zero);
    if Is_Zero then I := 0; return; end if;
    I := LST_Count(L*LMax2+S*LMax1+T)+(X-T)*(T-S+1)+(B-S)+1;  -- Indx(S,B,T,X,L)
  end Indx2;

  ---------------
  --- testing ---
  ---------------

  procedure ShowLM(N: in String; L1,L2,L3,M1,M2,M3: in Integer) is
    --- for testing Prod
  begin
    Show0(N & "(" & Strng(L1) & "," & Strng(M1) & ")",False);
    Show0(" * (" & Strng(L2) & "," & Strng(M2) & ")",False);
    Show0(" --> (" & Strng(L3) & "," & Strng(M3) & ")",True);
  end ShowLM;

  procedure CheckLM(N: String; L1,L2,L3,M1,M2,M3: Integer) is
    --- for testing
    LMin3: constant Integer := abs(L1-L2);
    LMax3: constant Integer := L1+L2;
    LSum: constant Integer := L1+L2+L3;
    MSum: constant Integer := M1+M2+M3;
  begin
    if abs(M1)>L1 then
      Show2(N & " invalid:  |M1|>L1 ",M1,L1);
    end if;
    if abs(M2)>L2 then
      Show2(N & " invalid:  |M2|>L2 ",M2,L2);
    end if;
    if abs(M3)>L3 then
      Show2(N & " invalid:  |M3|>L3 ",M3,L3);
    end if;
    if MSum/=0 then
      Show1(N & " invalid:  MSum/=0 ",MSum);
    end if;
    if L3<LMin3 then
      Show2(N & " invalid: L3<LMin3 ",L3,LMin3);
    end if;
    if L3>LMax3 then
      Show2(N & " invalid: L3>LMax3 ",L3,LMax3);
    end if;
    if M1=0 and then M2=0 and then (LSum mod 2)=1 then
      Show1(N & " invalid: LSum odd ",LSum);
    end if;
  end CheckLM;

end Regges;
