with Ada.Text_IO, ScalInfo, Logicals, Ints, Rationals, Flts;
use Ada.Text_IO, ScalInfo, Logicals, Ints, Rationals, Flts;
with Recycling, Vectors, Polynoms;

pragma Elaborate_All (Ada.Text_IO,ScalInfo,Logicals,Ints,Rationals);
pragma Elaborate_All (Recycling,Vectors);

generic

  Trunc: in Boolean;
  FDeg: in Natural;   --- max degree
  Rad: in Radius;

  -------------------- begin_include scalars Scalar

  type Scalar is private; --- assuming Real

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>; -- S1 constains S2
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>;
  with function Sup(S: Scalar) return Flt is <>;

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  --- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  -------------------- end_include scalars

  with package SV is new Vectors (Component => Scalar);
  with package Poly is new Polynoms (Scalar => Scalar, SV => SV);

package TCosSins1 is --- analytic cos series and sin series in 1 variable
                     --- trimmed-down version of CosSins1 with changes:
  --- Trunc and Rad are package args, errors are norms, FDeg=0 is treated like Scalar

  EDeg: constant Natural := 2*FDeg;
  subtype FFreq is Integer range 0 .. FDeg;
  subtype EFreq is Integer range 0 .. EDeg;
  subtype SPoly1 is Poly.Polynom1;
  subtype SVector is SV.Vector;
  subtype FPoly1 is SVector(FFreq);

  NoErrs: constant Boolean := Trunc or FDeg=0;
  Errs:   constant Boolean := not NoErrs;
  CutEDeg: constant Integer := Choose(NoErrs,-1,EDeg);
  subtype CutEFreq is Integer range 0 .. CutEDeg;
  subtype Errs1 is RadVec(CutEFreq);

  type CosSin1 is
    record
      P: Parity;  -- 0: even  1: odd. P=0 is assumed if FDeg=0
      C: FPoly1;  -- cos(K.) if P=0,  sin(K.) if P=1
      E: Errs1;   -- higher order frequency error norms
    end record;   -- note: do not use P=1 when FDeg=0

  Default_Parity: Parity := 0;
  procedure Show_C1_Param(All_Param: in Boolean := True);
  function CosSin(P: Parity) return String;                 --- "Cos" or "Sin"
  procedure XPowers(X: in Scalar; V: in out SVector);       --- V(N)=X^N
  procedure CSPowers(X: in Scalar; C,S: in out SVector);    --- C(I)=Cos(I*X), S(I)=Sin(I*X)

  --- basic
  function IsZero(P: CosSin1) return Boolean;
  procedure Parities(P: in CosSin1; Even,Odd: out Boolean); --- Even=Odd=True means P=0
  function IsSharp(P: CosSin1) return Boolean;              --- polynomial with sharp coeffs
  function IsPoly(P: CosSin1) return Boolean;
  function "="(P1,P2: CosSin1) return Boolean;
  procedure FillZero(P: in out CosSin1);                    --- does not set parity
  procedure SetZero(Par: in Parity; P: in out CosSin1);
  procedure SetZero(P: in out CosSin1);                     --- sets default parity
  procedure Copy(P1: in CosSin1; P2: in out CosSin1);
  procedure Swap(P1,P2: in out CosSin1);

  --- sets
  function Center0(P: CosSin1) return Boolean;
  function Contains0(P: CosSin1) return Logical;
  procedure BallAt0(R: in Flt; Par: in Parity; P: in out CosSin1);
  procedure BallAt0(R: in Flt; P: in out CosSin1);       --- sets default radius and parity
  function BallAt0(R: Flt) return CosSin1;               --- sets default radius and parity
  procedure ToErr(P: in out CosSin1);
  procedure ToErr(P1: in CosSin1; P2: in out CosSin1);
  function ToErr(P: CosSin1) return CosSin1;
  procedure Center(P: in out CosSin1);
  procedure Center(P1: in CosSin1; P2: in out CosSin1);
  function Center(P: CosSin1) return CosSin1;
  procedure ModCenter(P: in out CosSin1);
  procedure ModCenter(P1: in CosSin1; P2: in out CosSin1);
  function ModCenter(P: CosSin1) return CosSin1;            --- zero center
  procedure SplitErr(P: in CosSin1; PC,PE: in out CosSin1);
  procedure ErrMult(R: in Radius; P: in out CosSin1);

  --- basic arithmetic
  procedure Neg(P: in out CosSin1);
  procedure Neg(P1: in CosSin1; P2: in out CosSin1);
  function "-"(P: CosSin1) return CosSin1;
  procedure Add(I: in Integer; P: in out CosSin1);
  procedure Add(S: in Scalar; P: in out CosSin1);
  procedure Add(P1: in CosSin1; P2: in out CosSin1);
  procedure Sum(P1,P2: in CosSin1; P3: in out CosSin1);
  function "+"(P1,P2: CosSin1) return CosSin1;
  procedure Sub(P1: in CosSin1; P2: in out CosSin1);
  procedure Diff(P1,P2: in CosSin1; P3: in out CosSin1);
  function "-"(P1,P2: CosSin1) return CosSin1;
  procedure Mult(R: in Flt; P: in out CosSin1);
  procedure Prod(R: in Flt; P1: in CosSin1; P2: in out CosSin1);
  function "*"(R: Flt; P: CosSin1) return CosSin1;
  procedure AddProd(R: in Flt; P1: in CosSin1; P2: in out CosSin1);
  procedure Mult(Q: in Rational; P: in out CosSin1);
  procedure Mult(S: in Scalar; P: in out CosSin1);
  procedure Prod(S: in Scalar; P1: in CosSin1; P2: in out CosSin1);
  function "*"(S: Scalar; P: CosSin1) return CosSin1;
  procedure AddProd(S: in Scalar; P1: in CosSin1; P2: in out CosSin1);
  procedure Mult(P1: in CosSin1; P2,Tmp: in out CosSin1);
  procedure Mult(P1: in CosSin1; P2: in out CosSin1);
  procedure Prod(P1,P2: in CosSin1; P3: in out CosSin1);
  function "*"(P1,P2: CosSin1) return CosSin1;
  procedure AddProd(P1,P2: in CosSin1; P3: in out CosSin1);
  procedure AddProd(P1,P2: in CosSin1; P3,Tmp: in out CosSin1);
  procedure SubProd(P1,P2: in CosSin1; P3: in out CosSin1);
  procedure SubProd(P1,P2: in CosSin1; P3,Tmp: in out CosSin1);
  procedure Div(R: in Flt; P: in out CosSin1);
  procedure Div(S: in Scalar; P: in out CosSin1);
  procedure Quot(P1: in CosSin1; R: in Flt; P2: in out CosSin1);
  function "/"(P: CosSin1; R: Flt) return CosSin1;

  --- conversion and i/o
  function Get_Precision(P: CosSin1) return Positive;
  procedure Assign(I: in Integer; P: in out CosSin1);
  procedure Assign(Q: in Rational; P: in out CosSin1);
  procedure Assign(R: in Flt; P: in out CosSin1);
  procedure Assign(S: in Scalar; P: in out CosSin1);
  procedure Show(N: in String; P: in CosSin1; Cut: in Radius; Hide0: in Boolean := True);
  procedure Show1(N: in String; P: in CosSin1; Hide0: in Boolean := True);
  procedure Show2(N: in String; P1,P2: in CosSin1; Hide0: in Boolean := True);
  --- Caution: Rad is never Put/Get
  --- Simple means Cossin is Put/Get as Scalar.
  procedure Put(F: in File_Type; P: in CosSin1; Simple: in Boolean := False);
  procedure Get(F: in File_Type; P: in out CosSin1; Simple: in Boolean := False);
  procedure Write(FileName: in String; P: in CosSin1; Simple: in Boolean := False);
  procedure Read(FileName: in String; P: in out CosSin1; Simple: in Boolean := False);

  --- other
  procedure Simple_Random(P: in out CosSin1);
  procedure Simple_Random(R: in Radius; Par: in Parity; P: in out CosSin1);
  function MaxNorm(P: CosSin1) return Radius;
  procedure Norm(P: in CosSin1; S: in out Scalar);
  function Norm(P: CosSin1) return Scalar;
  function Inf(P: CosSin1) return Flt;
  function Sup(P: CosSin1) return Flt;

  procedure Average(P: in CosSin1; S: in out Scalar);
  function Average(P: CosSin1) return Scalar;
  procedure SetAverage0(P: in out CosSin1);
  procedure Coeff(I: in Natural; P: in CosSin1; S: in out Scalar);
  function Coeff(I: in Natural; P: CosSin1) return Scalar;
  procedure AddCoeff(I: in Natural; S: in Scalar; P: in out CosSin1);
  procedure AddErr(I: in Natural; R: in Radius; P: in out CosSin1);
  procedure AddBall(R: in Radius; P: in out CosSin1);
  function MaxVal(E: Errs1) return Radius;
  procedure LiftE0(E: in out Errs1);
  procedure ValAt0(P: in CosSin1; S: in out Scalar);
  procedure Val(P: in CosSin1; X: in Scalar; Y: in out Scalar; Tmp1,Tmp2: in out FPoly1);
  procedure Val(P: in CosSin1; X: in Scalar; Y: in out Scalar);
  function Val(P: CosSin1; S: Scalar) return Scalar;
  procedure TranslatePi(P: in out CosSin1);
  procedure TranslatePi(P1: in CosSin1; P2: in out CosSin1);
  procedure NegTranslatePi(P1: in CosSin1; P2: in out CosSin1);
  procedure SymmetrizePi(SP: in Parity; P: in out CosSin1; Check: in Boolean := False);
  procedure ProjEvenFreq(P: in out CosSin1);
  procedure ProjOddFreq(P: in out CosSin1);
  procedure ProjPar(Par: in Parity; P: in out CosSin1; Check: in Boolean := False);
  procedure ProjConst(P: in out CosSin1);

  procedure Der(P: in out CosSin1);
  procedure Der(P1: in CosSin1; P2: in out CosSin1);
  procedure AntiDer(P: in out CosSin1);
  procedure AntiDer(P1: in CosSin1; P2: in out CosSin1);
  procedure Integral(P: in CosSin1; X0,X: in Scalar; Y: in out Scalar);
  procedure InvHelm(A,B: in Scalar; Del: in Radius; P: in out CosSin1; OddFreq: in Boolean := True);

  ------------------ useful in TSperhics
  type Fourier1 is
    record
      E: CosSin1; -- even (cos)
      O: CosSin1; -- odd (sin)
    end record;

  procedure AddProd(S: in Scalar; P1: in CosSin1; P2: in out Fourier1);
  procedure Val(P: in Fourier1; X: in Scalar; Y: in out Scalar);

  --- numerical
  procedure NumTrunc(Siz: in Natural; P: in out CosSin1);
  procedure NumTrunc(Siz: in Natural; P1: in CosSin1; P2: in out CosSin1);
  function NumVal(P: CosSin1; T: Flt) return Flt;
  function NumVal(P: Fourier1; T: Flt) return Flt;
  procedure PutPlot(F: in File_Type; P: in CosSin1; Steps: in Positive := 64; PutX: in Boolean := False);
  function GuessRho(P: CosSin1) return Radius;

  SRad:         constant Scalar  := Scal(Rad);
  LastRadPower: constant Natural := IMax(0,2*EDeg);
  --- rounded up
  InvRad: constant Radius := One/Rad;
  RadPower: constant RadVec := PowersOf(Rad,LastRadPower);
  InvRadPower: constant RadVec := PowersOf(Rad,LastRadPower);

  EDegOdd: constant Integer := EDeg+(EDeg mod 2)-1;

  -----------------------------------------------------
  -----------------------------------------------------

  package PAlloc is new Recycling (Item => CosSin1);
  subtype LP_Pointer is PAlloc.LD_Pointer;
  PAC: PAlloc.Controlled_Protected_Data;
  PPool: PAlloc.Protected_Data renames PAC.P;

private

  subtype CNorms is RadVec(FFreq);

  procedure Show_C1_Rounding;

  function Is_Zero(P: Parity; C: FPoly1) return Boolean;
  function Is_Zero(E: Errs1) return Boolean;
  procedure CCProd2(P1P,P2P: in Parity; C1,C2: in FPoly1; A1,A2: in CNorms; P3: in out CosSin1);
  procedure CEProd2(P1P: in Parity; A1: in CNorms; E2: in Errs1; E3: in out Errs1);
  procedure EEProd2A(E1,E2: in Errs1; E3: in out Errs1);
  procedure EEProd2B(E1,E2: in Errs1; E3: in out Errs1);
  procedure EEProd2(E1,E2: in Errs1; E3: in out Errs1) renames EEProd2B;
  procedure Norms(P: in Parity; C: in FPoly1; A: out CNorms);
  procedure Product(P1,P2: in CosSin1; P3: in out CosSin1);
  function QuasiDeg(C: FPoly1; P: Parity := 0) return Natural;
  function QuasiDeg(E: Errs1) return Natural;
  procedure ValAt0(P: in CosSin1; S: in out Scalar; D: in Natural; Is_Poly: in Boolean);

  Not_Trunc: constant Boolean := not Trunc;
  SZero:      constant Scalar := Scal(Zero);
  SHalf:      constant Scalar := Scal(Half);
  SOne:       constant Scalar := Scal(One);
  STwo:       constant Scalar := Scal(Two);

end TCosSins1;
