with Rationals, Flts;
use Rationals, Flts;

pragma Elaborate_All (Rationals,Flts);

generic

  Trunc: in Boolean;
  Alpha: in Rational;
  type Scalar is private;

  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>;
  with function Sup(S: Scalar) return Flt is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;

package TSpherics_Aux is --- higher order bounds etc.

  --- for wave equation, periodic solutions
  procedure Wave_Del(L: in Positive; Del: out Radius; Tmp,Tmp1,Tmp2: in out Scalar);
  procedure Wave_Eps(L: in Positive; Eps,Tmp: in out Scalar);
  procedure Make_Wave_Del(Del: out RadVec);
  function Make_Wave_Del(SizeE: Positive) return RadVec;

  --- for plate equation, periodic solutions
  procedure Plate_Del(L: in Positive; Del: out Radius);

  --- for plate equation, quasiperiodic solutions
  procedure Make_Beta(I: in Integer; J: in Positive; Beta: in out Scalar);
  function Make_Beta(I: Integer; J: Positive) return Scalar;
  function Inf(Q: Rational) return Flt;
  function Dio_Const(J: Positive; N: Natural) return Radius;
  procedure QP_HiSigDel(Beta: in Scalar; L,M: in Positive; N: in out Positive; Sig,Del: out Radius);
  function SupAbs(S: Scalar) return Radius;
  function InfAbs(S: Scalar) return Radius;
  procedure QP_SigDel(Beta: in Scalar; L,M: in Positive; Sig,Del: out RadVec);
  procedure QP_Lam(I: in Integer; J,NMin: in Positive; Lam: out RadVec);

  --- some usful constants
  AlphaP:    constant Positive := INum(Alpha);
  AlphaQ:    constant Positive := IDen(Alpha);
  Alpha2:    constant Rational := Alpha*Alpha;
  SAlpha:    constant Scalar   := Scal(Alpha);
  SAlpha2:   constant Scalar   := Scal(Alpha*Alpha);
  SInvAlpha: constant Scalar   := Scal(Inv(Alpha));

end TSpherics_Aux;
